class StepperConfigurationError(Exception):
    '''
    Raised when error in stepper configuration found.
    '''


class AmmoFileError(Exception):
    '''
    Raised when failed to read ammo file properly.
    '''


class StpdFileError(Exception):
    '''
    Raised when failed to read stpd file properly.
    '''


class DiskLimitError(Exception):
    pass
